/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21 decembre 2006
   derniere modification le 02/07/2007 */


#ifndef __SCECAMERA_H__
#define __SCECAMERA_H__


/* constantes propres a Cam */
/* modes de la camera */
#define SCE_CAMERA_FPS 1
#define SCE_CAMERA_TPS 2
/* types de camera */
#define SCE_CAMERA_2D 1
#define SCE_CAMERA_3D 2



typedef struct sce_sprojmatrix SCE_SProjMatrix;
struct sce_sprojmatrix
{
    int w, h;
    float ratio;
    SCE_SInertVar angle;
    float near, far;
};


typedef struct sce_sviewport SCE_SViewport;
struct sce_sviewport
{
    int x, y;
    int w, h;
}; 


/* structure d'une camera */
typedef struct sce_scamera SCE_SCamera;
struct sce_scamera
{
    int type;           /* type de la camera (2D, 3D, ...) */
    
    /* matrice de projection */
    SCE_SProjMatrix projmatrix;
    
    SCE_TVector3 p;     /* position */
    SCE_TVector3 v;     /* vecteur de vue */
    SCE_TVector3 i;     /* inclinaison */
    
    int affect_p;       /* affectation de la position
                           en fonction du vecteur de vue */
    
    SCE_TVector3 fix;   /* point de fixation (TPS uniquement) */
    
    /* viewport de la camera */
    SCE_SViewport viewport;
    
    /* mode de la camera */
    int mode;
    
    /* booleen, true: utilisation de l'inertie */
    int use_rot_inert;
    int use_mov_inert;
    
    /* booleen, true: mouvements/rotations additionnels */
    int use_mov_add;
    int use_rot_add;
    
    /* variables d'inertie */
    SCE_SInertVar rotx, rotz, movx, movy, movz, radius;
    
    /* variable foireuse pour empecher la rotation verticale de merder */
    float back_rotx;
};



/* initialise une camera a ses valeurs par defaut */
void SCE_Camera_Init(SCE_SCamera*, int, int, int);


/* defini si les deplacements doivent etre relatifs au vecteur de vue */
void SCE_Camera_MoveFollowView(SCE_SCamera*, int);

/* defini la zone de l'ecran ou la camera sera representee */
void SCE_Camera_SetViewport(SCE_SCamera*, int, int, int, int);


/* defini l'angle de vision */
void SCE_Camera_SetVision(SCE_SCamera*, float);

/* defini le taux d'inertie de l'angle de vision */
void SCE_Camera_SetVisionInert(SCE_SCamera*, float, int);

/* defini le ratio... */
void SCE_Camera_SetRatio(SCE_SCamera*, float);

/* defini la valeur "proche" (near) */
void SCE_Camera_SetNear(SCE_SCamera*, float);

/* defini la valeur "eloignee" (far) */
void SCE_Camera_SetFar(SCE_SCamera*, float);


/* defini les valeurs de l'inertie du mouvement de la camera */
void SCE_Camera_SetMoveInert(SCE_SCamera*, float, int);

/* defini les valeurs de l'inertie de la rotation de la camera */
void SCE_Camera_SetRotateInert(SCE_SCamera*, float, int);

/* defini le rayon entre le point de fixation et la camera (TPS uniquement) */
void SCE_Camera_SetRadius(SCE_SCamera*, float);

/* defini le taux d'inertie du rayon */
void SCE_Camera_SetRadiusInert(SCE_SCamera*, float, int);

/* defini le mode de la camera */
void SCE_Camera_SetMode(SCE_SCamera*, unsigned int);

/* utilisation de l'inertie de rotation: oui, non */
void SCE_Camera_UseRotInert(SCE_SCamera*, int);

/* utilisation de l'inertie de mouvement: oui, non */
void SCE_Camera_UseMovInert(SCE_SCamera*, int);



/* defini le point de fixation de la camera
   (calculs adaptes en fonction du mode de la camera) */
void SCE_Camera_SetPointView(SCE_SCamera*, SCE_TVector3);


/* defini les prochaines rotations qu'effectuiera la camera specifiee
   (rotation adaptives en fonction du mode de la camera) */
void SCE_Camera_Rotate(SCE_SCamera*, float, float);


/* defini le prochain traget de la camera (avant/arriere ; latteral) */
void SCE_Camera_Move(SCE_SCamera*, float, float, float);

/* version vectorielle */
void SCE_Camera_Movev(SCE_SCamera*, SCE_TVector3);


/* calcul les nouvelles coordonnees de la camera specifiee */
void SCE_Camera_Compute(SCE_SCamera*);

/* defini la camera comme camera active */
void SCE_Camera_Use(SCE_SCamera*);


#endif /* guard */
