/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21 decembre 2006
   derniere modification le 02/07/2007 */


/**

IMPORTANT !! :
___
\p---> permettre d'avoir une matrice de projection par camera

**/


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEMath.h>
#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEInert.h>

#include <SCE/core/SCECMatrix.h>

#include <SCE/interface/SCECamera.h>


/* revise le 02/07/2007 */
void SCE_Camera_Init(SCE_SCamera *cam, int type, int w, int h)
{
    /* mode par defaut: point de fixation */
    cam->mode = SCE_CAMERA_TPS;
    
    cam->type = type;
    
    /* viewport par defaut */
    cam->viewport.x = cam->viewport.y = 0;
    cam->viewport.w = w;
    cam->viewport.h = h; /* NOTE: dimensions de l'ecran */
    
    cam->projmatrix.w = w;
    cam->projmatrix.h = h;
    cam->projmatrix.ratio = (float)w/h;
    SCE_Inert_Init(&cam->projmatrix.angle);
    cam->projmatrix.angle.var = cam->projmatrix.angle.real = 65.0f;
    cam->projmatrix.near = 0.1f;
    cam->projmatrix.far = 1000.0f;
    
    SCE_Vector3_Init(cam->p, 0.0f);
    SCE_Vector3_Init(cam->v, 0.0f);
    SCE_Vector3_Init(cam->i, 0.0f);
    SCE_Vector3_Init(cam->fix, 0.0f);
    
    /* la camera se deplacera coolement par defaut :P */
    cam->affect_p = SCE_TRUE;
    
    cam->back_rotx = 0.0f;
    
    /* inclinaison: verticale Z par defaut */
    cam->i[2] = 1.0f;
    
    /* point de vue par defaut: 0,0,0 */
    
    /* utilisation de l'inertie: non par defaut */
    cam->use_rot_inert = cam->use_mov_inert = 0;
    
    /* utilisation des deplacements additionnels, non par defaut */
    cam->use_mov_add = cam->use_rot_add = 0;
    
    /* initialisation des variables d'inertie */
    SCE_Inert_Init(&cam->rotx);
    SCE_Inert_Init(&cam->rotz);
    SCE_Inert_Init(&cam->movx);
    SCE_Inert_Init(&cam->movy);
    SCE_Inert_Init(&cam->movz);
    
    /* distance par rapport au point de fixation par defaut */
    SCE_Inert_Init(&cam->radius);
    cam->radius.var = cam->radius.real = 8.0f;
    
    /* les mouvements ne sont pas cumulatifs */
    SCE_Inert_Accum(&cam->movx, 0);
    SCE_Inert_Accum(&cam->movy, 0);
    SCE_Inert_Accum(&cam->movz, 0);
}



/* ajoute le 27/06/2007 */
void SCE_Camera_MoveFollowView(SCE_SCamera *cam, int follow)
{
    cam->affect_p = follow;
}

void SCE_Camera_SetViewport(SCE_SCamera *cam, int x, int y, int w, int h)
{
    cam->viewport.x = x;
    cam->viewport.y = y;
    cam->viewport.w = w;
    cam->viewport.h = h;
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetVision(SCE_SCamera *cam, float val)
{
    SCE_Inert_Set(&cam->projmatrix.angle, val);
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetVisionInert(SCE_SCamera *cam, float a, int b)
{
    SCE_Inert_SetParams(&cam->projmatrix.angle, a, b);
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetRatio(SCE_SCamera *cam, float val)
{
    cam->projmatrix.ratio = val;
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetNear(SCE_SCamera *cam, float val)
{
    cam->projmatrix.near = val;
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetFar(SCE_SCamera *cam, float val)
{
    cam->projmatrix.far = val;
}

void SCE_Camera_SetMoveInert(SCE_SCamera *cam, float inert, int np)
{
    SCE_Inert_SetParams(&cam->movx, inert, np);
    SCE_Inert_SetParams(&cam->movy, inert, np);
}

void SCE_Camera_SetRotateInert(SCE_SCamera *cam, float inert, int np)
{
    SCE_Inert_SetParams(&cam->rotx, inert, np);
    SCE_Inert_SetParams(&cam->rotz, inert, np);
}

/* revise le 01/07/2007 */
void SCE_Camera_SetRadius(SCE_SCamera *cam, float radius)
{
    SCE_Inert_Set(&cam->radius, radius);
}

/* ajoute le 02/07/2007 */
void SCE_Camera_SetRadiusInert(SCE_SCamera *cam, float a, int b)
{
    SCE_Inert_SetParams(&cam->radius, a, b);
}

void SCE_Camera_SetMode(SCE_SCamera *cam, unsigned int mode)
{
    SCE_TVector3 vec;
    
    cam->mode = mode;
    
    if(mode == SCE_CAMERA_TPS)
    {
        SCE_Vector3_Operator2v(vec, =, cam->v, -, cam->p);
        SCE_Vector3_Normalize(vec);
        SCE_Vector3_Operator3(cam->fix, =, cam->p, + ,
                              vec, *, SCE_Inert_Get(&cam->radius));
    }
    #ifdef __SCE_DEBUG__
    else if(mode != SCE_CAMERA_FPS)
    {   
        /* TODO: message d'erreur de merde */
        Logger_PrintMsg("- error : bad camera mode, set as default (fix).\n");
        cam->mode = SCE_CAMERA_TPS;
    }
    #endif
}

void SCE_Camera_UseRotInert(SCE_SCamera *cam, int use)
{
    cam->use_rot_inert = use;
}

void SCE_Camera_UseMovInert(SCE_SCamera *cam, int use)
{
    cam->use_mov_inert = use;
}



void SCE_Camera_SetPointView(SCE_SCamera *cam, SCE_TVector3 view)
{
    if(cam != NULL)
    {
        switch(cam->mode)
        {
            case SCE_CAMERA_FPS:
                SCE_Vector3_Operator2v(cam->v, =, view, -, cam->p);
                SCE_Vector3_Normalize(cam->v);
                
                break;
            
            case SCE_CAMERA_TPS:
                SCE_Vector3_Affectv(cam->v, view);
                break;
        }
    }
}


void SCE_Camera_Rotate(SCE_SCamera *cam, float z, float x)
{
    SCE_Inert_Set(&cam->rotx, x); /* haut/bas */
    SCE_Inert_Set(&cam->rotz, z); /* gauche/droite */
}


void SCE_Camera_Move(SCE_SCamera *cam, float x, float y, float z)
{
    SCE_Inert_Set(&cam->movx, x);   /* front */
    SCE_Inert_Set(&cam->movy, y);   /* lateral */
    SCE_Inert_Set(&cam->movz, z);   /* up/down */
}

void SCE_Camera_Movev(SCE_SCamera *cam, SCE_TVector3 v)
{
    SCE_Inert_Set(&cam->movx, v[0]);
    SCE_Inert_Set(&cam->movy, v[1]);
    SCE_Inert_Set(&cam->movz, v[2]);
}


/* revise le 28/06/2007 */
void SCE_Camera_Compute(SCE_SCamera *cam)
{
    SCE_TVector3 vec;
    float tmp;
    
    if(cam != NULL)
    {
        /* calcul de l'inertie si utilisee,
           sinon, simple calcul sans inertie */
        
        if(cam->use_rot_inert)
        {
            SCE_Inert_ComputeVar(&cam->rotx);
            SCE_Inert_ComputeVar(&cam->rotz);
        }
        else
        {
            SCE_Inert_Null(&cam->rotx);
            SCE_Inert_Null(&cam->rotz);
        }
        
        /* on empeche les variables de devenir trop grosses */
        while(cam->rotx.var >= 360.0)
        {
            SCE_Inert_Operator(cam->rotx, -=, 360.0f);
            cam->rotx.var -= 360.0f;
            cam->back_rotx -= 360.0f;
        }
        while(cam->rotx.var <= -360.0)
        {
            SCE_Inert_Operator(cam->rotx, +=, 360.0f);
            cam->rotx.var += 360.0f;
            cam->back_rotx += 360.0f;
        }
        
        while(cam->rotz.var >= 360.0)
        {
            SCE_Inert_Operator(cam->rotz, -=, 360.0f);
            cam->rotz.var -= 360.0f;
            cam->back_rotx -= 360.0f;
        }
        while(cam->rotz.var <= -360.0)
        {
            SCE_Inert_Operator(cam->rotz, +=, 360.0f);
            cam->rotz.var += 360.0f;
            cam->back_rotx += 360.0f;
        }
        /***/
        
        /* calcul de l'inertie de l'angle de vision */
        SCE_Inert_ComputeVar(&cam->projmatrix.angle);
        
        /* calcul de l'inertie du rayon */
        SCE_Inert_ComputeVar(&cam->radius);
        
        /* verification de l'angle... */
        if( (cam->back_rotx < 90.0f && SCE_Inert_Get(&cam->rotx) >= 90.0f) ||
            (cam->back_rotx > -90.0f && SCE_Inert_Get(&cam->rotx) <= -90.0f) ||
            
            (cam->back_rotx > 90.0f && SCE_Inert_Get(&cam->rotx) <= 90.0f) ||
            (cam->back_rotx < -90.0f && SCE_Inert_Get(&cam->rotx) >= -90.0f) ||
            
            (cam->back_rotx > 270.0f && SCE_Inert_Get(&cam->rotx) <= 270.0f) ||
            (cam->back_rotx < -270.0f && SCE_Inert_Get(&cam->rotx) >= -270.0f)||
            
            (cam->back_rotx < 270.0f && SCE_Inert_Get(&cam->rotx) >= 270.0f) ||
            (cam->back_rotx > -270.0f && SCE_Inert_Get(&cam->rotx) <= -270.0f))
            cam->i[2] = -cam->i[2];
        
        cam->back_rotx = SCE_Inert_Get(&cam->rotx);
        
        if(cam->use_mov_inert)
        {
            SCE_Inert_ComputeVar(&cam->movx);
            SCE_Inert_ComputeVar(&cam->movy);
            SCE_Inert_ComputeVar(&cam->movz);
        }
        else
        {
            SCE_Inert_Null(&cam->movx);
            SCE_Inert_Null(&cam->movy);
            SCE_Inert_Null(&cam->movz);
        }
        
        switch(cam->mode)
        {
            /* mode 'fps' */
            case SCE_CAMERA_FPS:
                cam->v[0] = cam->v[2] = 0.0;
                cam->v[1] = 1.0;
                
                /* rotations */
                SCE_Vector3_RotateX(cam->v,
                                    cos(SCE_Inert_Get(&cam->rotx)*RAD),
                                    sin(SCE_Inert_Get(&cam->rotx)*RAD));
                SCE_Vector3_RotateZ(cam->v,
                                    cos(SCE_Inert_Get(&cam->rotz)*RAD),
                                    sin(SCE_Inert_Get(&cam->rotz)*RAD));
                
                if(cam->affect_p)
                {
                    SCE_Vector3_Operator2(cam->p, +=, cam->v, *,
                                          SCE_Inert_Get(&cam->movx) );
                    
                    /* construction du vecteur latteral */
                    vec[0] = -cam->v[1];
                    vec[1] = cam->v[0];
                    vec[2] = cam->i[0];
                    
                    SCE_Vector3_Normalize(vec);
                    
                    /* deplacement latteral */
                    SCE_Vector3_Operator2(cam->p, +=, vec, *,
                                          SCE_Inert_Get(&cam->movy) );
                }
                else
                {
                    cam->p[0] += SCE_Inert_Get(&cam->movx);
                    cam->p[1] += SCE_Inert_Get(&cam->movy);
                    cam->p[2] += SCE_Inert_Get(&cam->movz);
                }
                
                
                /* point de vue final */
                SCE_Vector3_Operator1v(cam->v, +=, cam->p);
                
                break;
            
            /* mode 'visualisation' */
            case SCE_CAMERA_TPS:
                
                /* deplacement du point de fixation */
                if(cam->affect_p)
                {
                    SCE_Vector2_Operator2v(vec, =, cam->fix, -, cam->p);
                    SCE_Vector2_Normalize(vec);
                    SCE_Vector2_Operator2(cam->fix, +=, vec, *,
                                          SCE_Inert_Get(&cam->movx));
                    
                    /* construction du vecteur latteral */
                    tmp = vec[0];
                    vec[0] = -vec[1];
                    vec[1] = tmp;
                    vec[2] = cam->i[0];
                    
                    SCE_Vector2_Operator2(cam->fix, +=, vec, *,
                                          SCE_Inert_Get(&cam->movy));
                }
                else
                {
                    cam->fix[0] += SCE_Inert_Get(&cam->movx);
                    cam->fix[1] += SCE_Inert_Get(&cam->movy);
                    cam->fix[2] += SCE_Inert_Get(&cam->movz);
                }
                
                /* preparation a la rotation */
                cam->p[0] = 0.0f;
                cam->p[1] = SCE_Inert_Get(&cam->radius);
                cam->p[2] = 0.0f;
                
                /* rotation de la position de la camera */
                SCE_Vector3_RotateX(cam->p,
                                    -cos(SCE_Inert_Get(&cam->rotx)*RAD),
                                    -sin(SCE_Inert_Get(&cam->rotx)*RAD));
                SCE_Vector3_RotateZ(cam->p,
                                    cos(SCE_Inert_Get(&cam->rotz)*RAD),
                                    sin(SCE_Inert_Get(&cam->rotz)*RAD));
                
                /* translation basee sur le point de fixation */
                SCE_Vector3_Operator1v(cam->p, +=, cam->fix);
                
                /* construction du vecteur de vue */
                SCE_Vector3_Operator2v(cam->v, =, cam->fix, -, cam->p);
                
                
                /* si le rayon est negatif, on inverse le vecteur de vue
                   (regarde vers l'exterieur de la "sphere") */
                if(SCE_Inert_Get(&cam->radius) < 0.0f)
                {
                    /* construction du vecteur de vue inverse */
                    SCE_Vector3_Operator1v(cam->v, =-, cam->v);
                }
                
                /* translation du vecteur de vue pour le transformer
                   en point de vue */
                SCE_Vector3_Operator1v(cam->v, +=, cam->p);
                
                break;
        }
        
        
        if(!cam->use_mov_add)
        {
            /* la camera n'utilise pas les mouvements additionnels,
               replacage des valeurs de mouvements a zero */
            SCE_Inert_Set(&cam->movx, 0.0f);
            SCE_Inert_Set(&cam->movy, 0.0f);
        }
        
        if(!cam->use_rot_add)
        {
            /* la camera n'utilise pas les rotations additionnels,
               replacage des valeurs de rotations a zero */
            SCE_Inert_Set(&cam->rotx, 0.0f);
            SCE_Inert_Set(&cam->rotz, 0.0f);
        }
    }
}


void SCE_Camera_Use(SCE_SCamera* cam)
{
    SCE_TVector3 vec;
    SCE_SProjMatrix *m;
    
    SCE_CViewport(cam->viewport.x, cam->viewport.y,
                  cam->viewport.w, cam->viewport.h);
    
    m = &cam->projmatrix;
    
    if(cam->type == SCE_CAMERA_2D)
    {
        /* on place d'abord la matrice de projection */
        SCE_CLoadIdentitySetMatrix(GL_PROJECTION);
        SCE_CSetOrthoMatrix(0.0f, /*m->w*/1.0f, 0.0f, /*m->h*/1.0f, m->near, m->far);
        
        /* puis la matrice de modelisation/visualisation */
        SCE_CLoadIdentitySetMatrix(GL_MODELVIEW);
        /* TODO: sais pas encore faire ça :D */
        /* SCE_CLook2Dv(cam->p, cam->v, cam->i); */
        SCE_Vector3_Operator1v(vec, =-, cam->p);
        SCE_CTranslateMatrixv(vec);
        SCE_CRotateMatrix(SCE_Inert_Get(&cam->rotz), 0.0f, 0.0f, 1.0f);
    }
    else
    {
        /* on place d'abord la matrice de projection */
        SCE_CLoadIdentitySetMatrix(GL_PROJECTION);
        SCE_CSetPerspectiveMatrix(SCE_Inert_Get(&m->angle), m->ratio,
                                  m->near, m->far);
        
        /* puis la matrice de modelisation/visualisation */
        SCE_CSetMatrixMode(GL_MODELVIEW);
        SCE_CLook3Dv(cam->p, cam->v, cam->i);
    }
}
