/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 28 fevrier 2007
   derniere modification le 08/07/2007 */


#include <stdlib.h>
#include <ctype.h>

/* mode debug trop lourd dans Valof et Strof
 * + flemme de changer ca...
 */
/*
#ifdef __SCE_DEBUG__
 #define DEBUG 1
#else
 #define DEBUG 0
#endif*/


#if (DEBUG)
 #include <stdio.h>
#endif

#include <SCE/utils/SCEError.h>
#include <SCE/utils/SCEMemory.h>

#include <SCE/utils/SCEString.h>


/* renvoie le nombre de chiffres que contient arg (ex. 126 = 3, 4598 = 4) */
static unsigned int SCE_String_lenof(long arg)
{
    unsigned int size=1;
    long nb=1;
    
    if(arg < 0)
    {
        arg = -arg;
        size++;
    }
    
    while(1)
    {
        nb = nb*10;
        if(arg >= nb)
            size++;
        else
            break;
    }
    
    return (size);
}


/* renvoie a puissance coeff */
static long SCE_String_puissance(long a, int coeff)
{
    int i=0;
    const long val=a;
    
    if(coeff == 0)
        return (1);
    
    if(coeff < 0)
    {
        a = -a;
        for(i=0; i>coeff+1; i--)
            a = val*a;
    }
    else
    {
        for(i=0; i<coeff-1; i++)
            a = val*a;
    }
    
    return (a);
}


long double SCE_String_Valof(const char *str)
{
    unsigned int i = 0;         /* compteur */
    long double res = 0.0;      /* valeur de retour */
    float coeff = 0.0f;
    int val=0;
    unsigned int str_ilen = 0;  /* longueur de la valeur entiere */
    unsigned int str_dlen = 0;  /* longueur de la valeur décimale */
    unsigned int unit_tmp = 0;
    
    /* calcul de la longueur de la chaine str -->
       longueur de la partie entiere */
    while(str[str_ilen] != '\0' &&
          str[str_ilen] != '\n' &&
          str[str_ilen] != '.')
    {
        str_ilen++;
    }
    /* --> longueur de la partie decimale (si elle existe uniquement) */
    if(str[str_ilen] == '.')
    {
        while(str[str_dlen+str_ilen+1] != '\0' &&
              str[str_dlen+str_ilen] != '\n')
        {
            str_dlen++;
        }
    }
    
    if(str[0] == '-' || str[0] == '+')
        val = 1;
    
    coeff = SCE_String_puissance( 10, (str_ilen-val) );
    
    /* calcul de la valeur entiere */
    for(i=val; i<str_ilen; i++)
    {
        unit_tmp = str[i] - '0';
        if(unit_tmp > 9)
        {
            Logger_PrintMsg("argument %d : '%c' is not a number\n", i, str[i]);
            
            /* on saute, on s'en occupe pas */
            unit_tmp = 0;
        }
        else
            coeff /= 10.0f;
        
        res += unit_tmp*coeff;
    }
    
    coeff = 1.0f;
    /* calcul de la valeur decimale */
    for(i=str_ilen+1; i<str_dlen+str_ilen+1; i++)
    {
        unit_tmp = str[i] - '0';
        if(unit_tmp > 9)
        {
            Logger_PrintMsg("argument %d : '%c' is not a number\n", i, str[i]);
            
            /* on saute, on s'en occupe pas */
            unit_tmp = 0;
        }
        else
            coeff /= 10.0f;
        
        res += unit_tmp*coeff;
    }
    
    if(str[0] == '-')
        res = -res;
    
    unit_tmp = 0;
    coeff = 0;
    str_ilen = 0;
    str_dlen = 0;
    
    return (res);
}


char* SCE_String_Strof(long double arg, unsigned int nd)
{
    char* str = NULL;
    double vf = 0.0;
    int dec = 0;
    int dec1 = 0;
    unsigned int i = 0, j = 0;
    unsigned int lenof_arg = 0, lenof_dec = 0;
    unsigned int div = 0;
    
    lenof_arg = SCE_String_lenof(arg);
    
    /* on ne garde que la valeur decimale */
    vf = arg - (long)arg;
    
    /* si arg a une valeur decimale : */
    if(vf != 0.0)
    {
        dec = vf*SCE_String_puissance(10, nd);
        dec1 = vf*SCE_String_puissance(10, nd+1);
        
        if(dec1 + 5 >= SCE_String_puissance(10, nd+1))
        {
            arg++;
            dec = 0;
        }
        else
        {
            if(dec1 - (dec*10) >= 5)
                dec++;
        }
        
        lenof_dec = SCE_String_lenof(dec);
        
        str = SCE_malloc(lenof_arg+nd+1);
        if(str == NULL)
        {
            Logger_LogSrc();
            return (NULL);
        }
        /* initialisation du contenu */
        memset(str, '\0', sizeof(*str) * (lenof_arg+nd+1));
        
        if(arg < 0)
        {
            str[0] = '-';
            arg = -arg;
            i++;
        }
        
        
        while(i<lenof_arg)
        {
            div = SCE_String_puissance(10, lenof_arg-i-1);
            
            str[i] = (arg/div) + '0';
            i++;
            
            arg = arg-((int)(arg/div) * div);
        }
        
        str[i] = '.';
        i++;
        
        if(lenof_dec < nd)
        {
            for(j=0; j<(nd-lenof_dec); j++)
            {
                str[i] = '0';
                i++;
            }
            lenof_dec = nd;
        }
        
        
        while(i<lenof_arg+lenof_dec+1)
        {
            div = SCE_String_puissance(10, lenof_dec-i+lenof_arg);
            
            str[i] = (dec/div) + '0';
            i++;
            
            dec = dec-((int)(dec/div) * div);
        }
        
        vf=0;
        dec=0;
        i=0;
        lenof_arg=0;
        lenof_dec=0;
        div=0;
    }
    else
    {
        str = SCE_malloc(lenof_arg+1);
        if(str == NULL)
        {
            Logger_LogSrc();
            return (NULL);
        }
        /* initialisation du contenu */
        memset(str, '\0', sizeof(*str) * (lenof_arg+1));
        
        if(arg < 0)
        {
            str[0] = '-';
            arg = -arg;
            i++;
        }
        
        while(i<lenof_arg)
        {
            div = SCE_String_puissance(10, lenof_arg-i-1);
            
            str[i] = (arg/div) + '0';
            i++;
            
            arg = arg-((int)(arg/div) * div);
        }
        
        vf=0;
        dec=0;
        i=0;
        lenof_arg=0;
        lenof_dec=0;
        div=0;
    }
    
    return (str);
}


unsigned int SCE_String_NChrInStr(const char *str, char c)
{
    size_t i;
    
    /* nombre d'occurrences */
    unsigned int nb = 0;
    
    size_t len = strlen(str);
    
    
    for(i=0; i<len; i++)
    {
        if(str[i] == c)
            nb++;
    }
    
    return (nb);
}


char* SCE_String_GetExt(char *str)
{
    size_t i = strlen(str);
    
    while(i != 0 && str[i] != '.')
        i--;
    
    if(i == 0 && str[0] != '.') /* sans extension... */
        return (NULL);
    
    return (&str[i]);
}


int SCE_String_Cmp(const char *str1, const char *str2, int cmp_case)
{
    size_t i;
    size_t len, len2;
    int equal = 1;
    
    /* on recupere la plus petite taille */
    len = strlen(str1);
    len2 = strlen(str2);
    
    if(len != len2)
        return (0);
    
    if(cmp_case)
    {
        for(i=0; i<len; i++)
        {
            if(str1[i] != str2[i])
            {
                equal = 0;
                break;
            }
        }
    }
    else
    {
        for(i=0; i<len; i++)
        {
            if(tolower(str1[i]) != tolower(str2[i]))
            {
                equal = 0;
                break;
            }
        }
    }
    
    return (equal);
}
